/*
   Module   : Harbour Mini Rules Engine

   author   : Frank Alleman
   version  : 1.0.0
   date     : 21-02-2014
*/

/* HMG include */
#include "hmg.ch"
#include "hbclass.ch"
/* end HMG include */

#define _DECLARATIVE
#define _LOGFILE
#define pRET	13
#define pLINEFEED	10

/*
   Usage       : Calculate requested fact
   
   Description:
   The mechanism works as follows.
   First the candidate rules (with fact := ...) are selected and placed into the agenda.
   The condition part of each rule of the agenda is evaluated, until a condition evaluates to true or 
   all conditione have been evaluated unsuccessfully.
   A condition can only be evaluated when all facts within that condition are known in the factset, 
   then the facts can be replaced by their values. Otherwise a fact is placed into the "varlist". 
   For each fact of the varlist the fact becomes a new goal an this function is called recursively.
   When no rules can be found or all condition evaluates to false, the user has to enter the value for 
   the fact manually.
   When an condition evaluatres to true, the action can be evaluated.
   The facts within the action have to be replaces by their values as well.
   When a fact can not be found, again that fact is placed into a varlist and for each fact of the varlist
   this fuction is called recursively .
      
   Parameters:
   cGoal    fact to be calculated 
   aRuleset array with selected rules 
   afactset array with facts (name/value pairs)
   nLevel   for internal use, the calling application only has to provide the first 3 parameters   
   Returnvalue:
   value    The calculated value
*/
function Evaluate(cGoal, aRuleset, aFactset, nLevel)
local aAgenda, cCondition, cAction, aVarlist, aWordlist, cbBlock, value := 0, n, m, pos, logfile, cStr, ar, len, retval
private cCodeString
public cLabelGoal, cFormValue
if pcount() = 3
   nLevel := 1
   len := len(aFactset)
endif
aAgenda := {}
CreateAgenda(cGoal, aRuleset, aAgenda)
for n := 1 to len(aAgenda)
	#ifdef _DECLARATIVE
		ar := {}
		cCondition := aAgenda[n, 1]
		aadd(ar, {{}, cCondition, "", ''})
		ParseCondition(ar[1])
		if EvaluateCondition(ar, aRuleset, aFactset, nLevel)
			exit
		endif
	#else
      aWordlist := {}
		aVarlist := {}
		cCondition := aAgenda[n, 1]
		CheckRule(cCondition, aVarlist, aWordlist, aFactset, "condition")
		for m := 1 to len(aVarlist)
         /* varlist contains facts that have to be resolved first */
			Evaluate(aVarlist[m], aRuleset, aFactset, nLevel + 1)
		next
		// condition resolvable
		ResolveRule(@cCondition, aWordlist, aFactset, "condition")	
		cCodeString := "{|__r,__f|" + cCondition + "}"
		cbBlock := &cCodeString.
      retval := eval(cbBlock, aRuleset, aFactset) 
      #ifdef _LOGFILE
         logfile := fopen("rules.log", 2)
         fseek(logfile, 0, 2)
         if retval
            cStr := ".T."
         else
            cStr := ".F."
         endif
         fwrite(logfile, "condition " + cCondition + " := " + cStr + chr(pRET) + chr(pLINEFEED))
         fclose(logfile)
      #endif
		if retval
			exit
		endif
	#endif
next
if n > len(aAgenda)
	/* ? "cCondition not resolvable" */
   cLabelGoal := cGoal

/* HMG include */
   load window RE0010
	RE0010.Activate()
   value := cFormValue
/* end HMG include */

/* Clipper include */
/*
   cls
	set cursor on
	oldcolor := setcolor("N/BG")
	@ 10, 2 clear to 13,77
	//@ 10, 2 to 13, 77 double
	setcolor(oldcolor)
 	// title
	title := "Invoeren feit"
	@ 10, 2 say replicate(" ", 77-2+1) color "W/B"
	@ 10, 2 + ((77-2+1 - len(title))/2) say title color "W/B"
	// teken veld
	value := space(22)
	@ 10+2,2+2 say padr(cGoal,30) color "N/BG" get value color "N/W,W+/N,,,N/W"
	read
*/
/* end Clipper include */

	do case
		case substr(value,1,4) = "ctod"
			value := ctod(substr(value,7,10))
		case substr(value,1,1) = '"'
			value := rtrim(value)
			value := substr(value,2,len(value)-2)
		case upper(substr(value,1,1)) $ "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
			value := rtrim(value)
		case substr(value,1,1) $ "-0123456789"
			value := val(value)
		case substr(value,1,3) = ".T."
			value := .T.
		case substr(value,1,3) = ".F."
			value := .F.
	endcase			
else
	/* condition resolved */
	aVarlist := {}
   aWordlist := {}
	cAction := aAgenda[n, 2]
	CheckRule(cAction, aVarlist, aWordlist, aFactset, "action")
	for m := 1 to len(aVarlist)
      /* varlist contains facts that have to be resolved first */
		Evaluate(aVarlist[m], aRuleset, aFactset, nLevel + 1)
	next
	/* action resolvable */
	ResolveRule(@cAction, aWordlist, aFactset, "action")
   pos := at("=", cAction)
   /* skip assignment part */
	cCodeString := "{|__r,__f|" + ltrim(substr(cAction, pos + 1)) + "}"
	cbBlock := &cCodeString.
	value := eval(cbBlock, aRuleset, aFactset)
	#ifdef _LOGFILE
		logfile := fopen("rules.log", 2)
		fseek(logfile, 0, 2)
		if valtype(value) = "C"
			cStr := value
		elseif valtype(value) = "N"
			cStr := ltrim(str(value))
		elseif valtype(value) = "D"
			cStr := dtoc(value)
		elseif valtype(value) = "L"
			cStr := if(value, ".T.", ".F.")
		endif
		fwrite(logfile, "action " + cAction + " := " + cStr + chr(pRET) + chr(pLINEFEED))
		fclose(logfile)
	#endif
endif	
if nLevel = 1
   // replace existing fact
   AddFact(aFactset, {cGoal, value})	
   ResetFactset(aFactset, len)
else   
   AddFact(aFactset, {cGoal, value})	
endif
return(value)

/* HMG include */
function enterfact_form_oninit()
   RE0010.label_1.value := cLabelGoal
Return(nil)

function enterfact_ok_onok()
   cFormValue := RE0010.text_1.value
   RE0010.release   
Return(nil)
/* End HMG include */

/*
   Goal:
   Creates agenda with rules that have to be evaluated.
   When the condition part of a rule evaluates to true, the action is executed and the value is returned;
   the remaining rules doesn't have to be evaluated any more. 
   Params:
   cGoal       the fact tht has to be calculated
   aRuleset    array with selected rules
   aAgenda     agenda with candidate rules to evaluate
   Returnvalue:
   aAgenda     agenda with candidate rules to evaluate
*/
function CreateAgenda(cGoal, aRuleset, aAgenda)
local n, len_aRuleset := len(aRuleset), len_cGoal := len(cGoal)
for n := 1 to len_aRuleset
	if substr(aRuleset[n, 2], 1, len_cGoal) == cGoal
	  if substr(aRuleset[n, 2], len_cGoal + 1, 1)$": "
			aadd(aAgenda, {aRuleset[n, 1], aRuleset[n, 2]})
		endif
	endif
next	
return(aAgenda)

/*
   Goal:
   Loads requested rules from the rulebase into the ruleset.
   Params:
   nClassId    classid 
   aRuleset    array with selected rules
   Returnvalue:
   aRuleset    array with selected rules
*/
function LoadRuleset(cLevel, aRuleset)
local old := select()
select 0
if cLevel == "Document" /* Document */
	use rsetdoc
	do while .not. eof()
		aadd(aRuleset, {rtrim(CONDITION), rtrim(ACTION)})
		skip
	enddo	
	use
else
	use rset index rsetb
   seek padr(cLevel, 30)
	do while .not. eof() .and. rtrim(rset->LEVEL) == cLevel
      aadd(aRuleset, {rtrim(CONDITION), rtrim(ACTION)})
		skip
	enddo	
	use
endif
select(old)
return(aRuleset)

/*
   Goal:
   Checks whether or not facts within the rulepart can be found in the factset.
   Params:
   cRulepart   the condition or action part of a rule
   aVarlist    list with facts that have not been found 
   aWordlist   the list to be filled with facts 
   aFactset    array with facts
   cOption     indicates whether rulepart concerns condition or action
   Returnvalue:
   aVarlist    list with facts that have not been found
*/
function CheckRule(cRulepart, aVarlist, aWordlist, aFactset, cOption)
local len_aWordlist, len_aFactset := len(aFactset)
ParseRule(cRulepart, aWordlist, cOption)
len_aWordlist := len(aWordlist)
for n:= 1 to len_aWordlist
	/* lookup var in aFactset */
	for m := 1 to len_aFactset
		if aWordlist[n, 1] == aFactset[m, 1]
         aWordlist[n, 2] := m
			exit
		endif
	next
	if m > len_aFactset
		p := ascan(aVarlist, {|x|x == aWordlist[n, 1]})
		if p = 0 .and. aWordlist[n, 1] != "__r" .and. aWordlist[n, 1] != "__f" 
			aadd(aVarlist, aWordlist[n, 1])
		endif
	endif		
next
return(aVarlist)

/*
   Goal:
   Replaces facts within a rulepart by their values.
   Params:
   cRulepart   the condition or action part of a rule
   aWordlist   the list to be filled with facts 
   aFactset    array with facts
   cOption     indicates whether rulepart concerns condition or action
*/
function ResolveRule(cRulepart, aWordlist, aFactset, cOption)
local cStr, pos, pos2, left, right, n, m, p := 0, q, searchstr, len_aWordlist, len_aFactset := len(aFactset)
len_aWordlist := len(aWordlist)
for n:= 1 to len(aWordlist)
	m := aWordlist[n, 2]
   /* assert(aWordlist[n, 1] == aFactset[m, 1]) */
   if m = 0
      for q := 1 to len(aFactset)
         if aWordlist[n, 1] == aFactset[q, 1]
            m := q
            exit
         endif   
	  next
   endif
   if m > 0
      cStr := aFactset[m, 2]
      if valtype(aFactset[m, 2]) = "C"
         cStr := '"' + aFactset[m, 2] + '"'
      elseif valtype(aFactset[m, 2]) = "N"
         cStr := ltrim(str(aFactset[m, 2]))
      elseif valtype(aFactset[m, 2]) = "D"
         cStr := 'ctod("' + dtoc(aFactset[m, 2]) + '")'
      elseif valtype(aFactset[m, 2]) = "L"
         cStr := if(aFactset[m, 2], ".T.", ".F.")
      endif
      if cOption == "action"
         // start after '='
         for p := 1 to len(cRulepart)
            if substr(cRulepart, p, 1) = "="
               exit
            endif
         next
      endif
      pos2 := p + 1
      do while .T.
         searchstr := substr(cRulepart, pos2)
         pos := at(aWordlist[n, 1], searchstr)
         if pos = 0
            exit
         else
            left := pos - 1
            right := pos + len(aWordlist[n, 1])
            if (left > 0 .and. upper(substr(searchstr,left,1))$"ABCDEFGHIJKLMNOPQRSTUVWXYZ.:_");
            .or. (right <= len(searchstr) .and. upper(substr(searchstr,right,1))$"ABCDEFGHIJKLMNOPQRSTUVWXYZ.:_")
               pos2 += pos
            else
               cRulepart := stuff(cRulepart, pos2+pos-1, len(aWordlist[n, 1]), cStr)
               exit
            endif
         endif	
      enddo
   endif
next
return(nil)

/*
   Goal:
   extracts facts from the rulepart and puts them into the wordlist
   Params:
   cRulepart   the condition or action part of a rule
   aWordlist   the list to be filled with facts 
   cOption     indicates whether rulepart concerns condition or action
*/
function ParseRule(cRulepart, aWordlist, cOption)
local cWord := "", m := 0, n, teller := 0, len_cRulepart := len(cRulepart)
if cOption == "action"
	// start after '='
	for m := 1 to len_cRulepart
		if substr(cRulepart, m, 1) = "="
			exit
		endif
	next
endif
m += 1
for n := m to len_cRulepart
	if (substr(cRulepart, n, 1) = '"')
		teller++
		if teller > 1
			teller := 0
		endif
	endif	
	if teller = 0
		if (asc(substr(cRulepart, n, 1)) > 64 .and. asc(substr(cRulepart, n, 1)) < 91) ;
		.or. (asc(substr(cRulepart, n, 1)) > 96 .and. asc(substr(cRulepart, n, 1)) < 123) ;
		.or. substr(cRulepart, n, 1) = "_" ; 
		.or. substr(cRulepart, n, 1) = ":" ;
		.or. (substr(cRulepart, n, 1) = "." .and. !(substr(cRulepart, n+1, 1) $ "0123456789")) ; 
		.or. (substr(cRulepart,n,1) $ "0123456789" .and. len(cWord) > 0) /* digit within string */
			cWord := cWord + substr(cRulepart, n, 1)
		else
			if len(cWord) > 0
            if substr(cWord, 1, 1) = "_"
					aadd(aWordlist, {cWord, 0})
				endif
				cWord := ""
			endif
		endif
	endif
next
if len(cWord) > 0
   if substr(CWord, 1, 1) = "_"
      aadd(aWordlist, {cWord, 0})
   endif   
endif
return(nil)

/*
   Goal:
   Retrieves value of requested fact from the factset
   Params:
   aFactset    array with facts
   cName       name of the fact to be retrieved
   Returnvalue:
   retval      value of the requested fact
*/
function GetFact(aFactset, cName)
local retval := "", n
for n := 1 to len(aFactset)
	if aFactset[n, 1] == cName
		retval := aFactset[n, 2]
		exit 
	endif
next
return(retval)

/*
   Goal:
   Adds fact to the factset
   Params:
   aFactset    array with facts
   cElem       fact that has to be added to array
   Returnvalue:
*/
function AddFact(aFactset, cElem)
local n, oldValue := nil 
n := ascan(aFactset, {|x|x[1] == cElem[1]})	
if n > 0
   oldValue := aFactset[n, 2]
	/* replace value */
	aFactset[n, 2] := cElem[2]
else
	/* add */
	aadd(aFactset, cElem)
endif
return(oldValue)

/*
   Goal:
   Deletes fact from the factset and resizes tha array
   Params:
   aFactset    array with facts
   cElem       fact that has to be added to array
   Returnvalue:
*/
function DelFact(aFactset, cElem)
local n
n := ascan(aFactset, {|x|x[1] == cElem})	
if n > 0
	/* delete fact */
	adel(aFactset, n)
	asize(aFactset, (len(aFactset) - 1))
endif
return(nil)

/*
   Goal:
   Resets the lenght of the ruleset to the given lenght.
   Params:
   aRuleset    array with rules
   len_old     the given lenght
   Returnvalue:
*/
function ResetRuleset(aRuleset, len_old)
local len := len(aRuleset)
do while len > len_old
	adel(aRuleset, len)
	len--
enddo 
asize(aRuleset, len_old)
return(nil)

/*
   Goal:
   Resets the lenght of the factset to the given lenght.
   Params:
   aRuleset    array with facts
   len_old     the given lenght
   Returnvalue:
*/
function ResetFactset(aFactset, len_old)
local len := len(aFactset)
do while len > len_old
	adel(aFactset, len)
	len--
enddo 
asize(aFactset, len_old)
return(nil)

/*
   Goal:
   Parses the condition part recursively and puts the elments into a tree (ar)
   Params:
   ar       array (tree) with elemenents
   Returnvalue:
*/
function ParseCondition(ar)
local nTellerString := 0, nTellerOpeningParenthesis := 0, nPosOpeningParenthesis, nPosClosingParenthesis, nPosOperand := 0, nPosLeft := 1, ;
len_cCondition, cCondition, len_ar1, n := 1, bClosingParenthesis := .F., bFunctie := .F., bNOT := .F., nTellerFunctionOpeningParenthesis := 0

/*
// ar points to array
// nTellerString > 0 means n is withing a string
// nTellerOpeningParenthesis is a teller for open parentheses 
// bClosingParenthesis .T. means that last processed conditionpart ended with a close parenthesis
// nPosOpeningParenthesis indicates beginposition first open parenthesis
// nPosClosingParenthesis indicates position last close parenthesis
// nPosOperand > 0 means that operand of .and. or .or. has to be processed from posision nPosOperand
// nPosLeft indicates startposition for investigating current conditionpart
// bFunctie .T. means that n is within a function
*/

cCondition := ar[2]
len_cCondition := len(cCondition)
do while n <= len_cCondition
	if (substr(cCondition, n, 1) = '"') /* begin string */
		nTellerString++
		if nTellerString > 1
			nTellerString := 0
		endif
	endif	
	if nTellerString = 0 /* at this point you are not wihin a string */
		if (substr(cCondition, n, 1) = "(") /* beginhaakje */
			if bFunctie = .T.
            nTellerFunctionOpeningParenthesis++
         else
				nTellerOpeningParenthesis++
				if nTellerOpeningParenthesis = 1 /* first beginhaakje */
					nPosOpeningParenthesis := n
				endif
			endif
		elseif (substr(cCondition, n, 1) = ")") /* eindhaakje */
			if bFunctie = .T.
            if nTellerFunctionOpeningParenthesis = 1
               bFunctie := .F.
            else
               nTellerFunctionOpeningParenthesis--
            endif
			else
				if nTellerOpeningParenthesis = 1 /* corresponding beginhaakje found */
					bClosingParenthesis := .T.
					nPosClosingParenthesis := n
               if bNOT
                  if n < len_cCondition
                     aadd(ar[1], {{}, ltrim(substr(cCondition, nPosLeft, nPosClosingParenthesis-nPosLeft+1)), "", ""})
                  else
                     if nPosLeft = 1
                        ar[4] := "REV"
                  		aadd(ar[1], {{}, ltrim(substr(cCondition, nPosOpeningParenthesis+1, nPosClosingParenthesis-nPosOpeningParenthesis-1)), "", ""})
                     else
		                  aadd(ar[1], {{}, ltrim(substr(cCondition, nPosLeft)), "", ""})
                     endif
                  endif
                  bNOT := .F.
               else   
                  aadd(ar[1], {{}, ltrim(substr(cCondition, nPosOpeningParenthesis+1, nPosClosingParenthesis-nPosOpeningParenthesis-1)), "", ""})
               endif   
					nTellerOpeningParenthesis := 0
					len_ar1 := len(ar[1])
					ParseCondition(ar[1, len_ar1])
					nPosOperand := 0
					nPosLeft:= n + 1
				else
					nTellerOpeningParenthesis--
				endif
			endif
		elseif upper(substr(cCondition, n)) = ".AND."  
         if nTellerOpeningParenthesis = 0 /* beginhaakje not encountered */
            bNOT := .F.
            if bClosingParenthesis = .T.
               /* rulepart at teh teft of nPosLeft has already been processed */
					len_ar1 := len(ar[1])
					ar[1, len_ar1, 3] := ".and."	
					nPosOperand := n + 5
					exit
				endif			
				aadd(ar[1], {{}, ltrim(substr(cCondition, nPosLeft, n-nPosLeft-1)), ".and.", ""}) 
				len_ar1 := len(ar[1])
				ParseCondition(ar[1, len_ar1])
				n += 4
				nPosOperand := n + 1
				nPosLeft := n + 1
			endif
		elseif upper(substr(cCondition, n)) = ".OR."  
			if nTellerOpeningParenthesis = 0 /* beginhaakje not encountered yet */
            		bNOT := .F.
				if bClosingParenthesis = .T.
			            /* rulepart at teh teft of nPosLeft has already been processed */
					len_ar1 := len(ar[1])
					ar[1, len_ar1, 3] := ".or."	
					nPosOperand := n + 4
					exit
				endif			
				aadd(ar[1], {{}, ltrim(substr(cCondition, nPosLeft, n-nPosLeft-1)), ".or.", ""}) 
				len_ar1 := len(ar[1])
				ParseCondition(ar[1, len_ar1])
				n += 3
				nPosOperand := n + 1
				nPosLeft := n + 1
			endif
		elseif upper(substr(cCondition, n)) = ".NOT."  
			if nTellerOpeningParenthesis = 0 /* beginhaakje not encountered yet */
	               	bNOT := .T.
				nPosLeft := n
			endif
			n += 4
		elseif upper(substr(cCondition, n)) = "CTOD"  
         if bFunctie = .T.
            /* you are already within a function */
         else
            bFunctie := .T.
            nTellerFunctionOpeningParenthesis := 0
         endif   
		elseif upper(substr(cCondition, n)) = "SUBSTR"  
			bFunctie := .T.
		elseif upper(substr(cCondition, n)) = "ISEMPTY"  
			bFunctie := .T.
		elseif upper(substr(cCondition, n)) = "ISNOTEMPTY"  
			bFunctie := .T.
		elseif upper(substr(cCondition, n)) = "GET"  
			bFunctie := .T.
		endif
	endif
	n++
enddo
if nPosOperand > 0 
	aadd(ar[1], {{}, ltrim(substr(cCondition, nPosOperand)), "", ""}) 
	len_ar1 := len(ar[1])
	ParseCondition(ar[1, len_ar1])
endif
return(nil)

/*
   Goal:
   Evaluates the conditip parts of the tree (ar).
   Params:
   ar       array (tree) with condition elements
   Returnvalue:
*/
function EvaluateCondition(ar, aRuleset, aFactset, nLevel)
local n, m, retval := .T., cCondition, aWordlist, aVarlist, cbBlock, cStr
private cCodeString
for n := 1 to len(ar)
	if empty(ar[n, 1])
		/* evaluate */
		cCondition := ar[n, 2]
		aVarlist := {} 
      aWordlist := {}
		CheckRule(cCondition, aVarlist, aWordlist, aFactset, "condition")
      /* varlist contains facts that have to be resolved first */
		for m := 1 to len(aVarlist)
			Evaluate(aVarlist[m], aRuleset, aFactset, nLevel + 1)
		next
		/* condition resolvable */
		ResolveRule(@cCondition, aWordlist, aFactset, "condition")	
		cCodeString := "{|__r,__f|" + cCondition + "}"
		cbBlock := &cCodeString.
		retval := eval(cbBlock, aRuleset, aFactset)
      #ifdef _LOGFILE
         logfile := fopen("rules.log", 2)
         fseek(logfile, 0, 2)
         if retval
            cStr := ".T."
         else
            cStr := ".F."
         endif
         fwrite(logfile, "condition " + cCondition + " := " + cStr + chr(pRET) + chr(pLINEFEED))
         fclose(logfile)
      #endif
   else
      /* adjusted 15-02-2013 */
      // if upper(substr(ar[n, 2], 1, 5)) == ".NOT." .and. substr(ltrim(substr(ar[n, 2], 6)),1,1) = '(' ;
      //    .and. upper(substr(ar[n, 1, 1, 2], 1, 5)) != ".NOT."
      if upper(substr(ar[n, 2], 1, 5)) == ".NOT." .and. ar[n, 4] == "REV"
         retval := if(EvaluateCondition(ar[n, 1], aRuleset, aFactset, nLevel), .F., .T.)
      else      
         retval := EvaluateCondition(ar[n, 1], aRuleset, aFactset, nLevel)	 
      endif   
	endif
	if retval .and. ar[n, 3] == ".or."
      /* when the left side evaluates to true, the right side of the .or. can be skipped */
		exit
	endif
	if retval = .F. .and. ar[n, 3] == ".and."
      /* when the left side evaluates to false, the right side of the .and. can be skipped */
		exit
	endif
next
return(retval)

function IsEmpty(exp)
return(empty(exp))

function IsNotEmpty(exp)
return(!empty(exp))

/* end-of-file */